#!/usr/bin/perl -w

# bf2pl.pl - A Brainfuck to Perl converter / interpreter
# Copyright (C) 2000 Gary LaRocco  (mr_rock@cydathria.com)
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

use Getopt::Long;

print "bf2pl Ver 1.00, Copyright (C) 2000 Gary LaRocco\n\n";
print "bf2pl comes with ABSOLUTELY NO WARRANTY. This is free\n";
print "software. As such, you are welcome to redistribute it\n";
print "under certain conditions. See the GNU General Public\n";
print "License for details.\n\n";

&SetOptions;

# Convert all files specified on the command line
for (@ARGV) { CvtFromFile($_); }

exit;

#----------------------------------------------------------------------

sub SetOptions {

    $byte    = 0;
    $classic = 0;
    $circ    = 0;
    $exec    = 1;
    $help    = 0;
    $opt     = 1;
    $pchk    = 0;
    $pmax    = 0;
    $target  = "perl";

    GetOptions ('byte!'    => \$byte,
                'classic!' => \$classic,
                'circ!'    => \$circ,
                'exec!'    => \$exec,
                'help'     => \$help,
                'opt!'     => \$opt,
                'pchk!'    => \$pchk,
                'pmax=i'   => \$pmax,
                'target=s' => \$target);

#   Classic mode sets a max array size of 30000 bytes,
#   allows only single byte unsigned values, and will
#   cancel circular mode..
    if ($classic) { 
        $byte = 1; 
        $circ = 0;
        $pmax = 30000;
    }


#   Initalize for the target language.
    $target = lc($target);
    for ($target) {
        /perl/ && do { &Init_for_Perl; last; }
    }


#   Display some help if there is nothing to process.
    &DisplayHelp if (($help) || ($#ARGV == -1));
}

#----------------------------------------------------------------------

sub ReadBFSource {
    my ($infile) = @_;
    my $bfcode;

    if (open (INFILE,$infile)) {
        while (<INFILE>) {
            $bfcode .= $_;
        }
        close(INFILE);
        return $bfcode;
    } else {
        warn "Can not open $infile for input\n";
        return 0;
    }
}
#----------------------------------------------------------------------

sub CvtFromFile {

    my ($infile) = @_;
    my $outfile;
    my $bfcode;
    my $cvtcode;

    ($outfile) = split /\./, $infile;
    $outfile .= $file_ext;

    $bfcode = ReadBFSource($infile);
    if ($bfcode) {
        if (open (OUTFILE,">$outfile")) {
            print "Converting $infile -> $outfile\n";
            $cvtcode = CvtBF($bfcode);
            print OUTFILE $cvtcode;
            close(OUTFILE);
        } else {
            warn "Can not open $outfile for output\n";
        }
        close(INFILE);

        if ($exec && $cvtcode && ($target eq 'perl')) {
            print "Executing $outfile\n\n";
            eval $cvtcode || die "$@\n";
        }
    }
}

#----------------------------------------------------------------------

sub CvtBF {

    my ($bfcode) = @_;
    my $cvtcode;
    my $indlevel;
    my $indstr   = '';
    my $cnt_p    = 0;
    my $cnt_a    = 0;
    my $c;
    my $prev_c;
    my $nesting  = 0;
    my $pchktxt;

#   Strip out all non Brainfuck characters
    $bfcode =~ s/[^\<\>\+\-\.\,\[\]\#\!]//g;

#   Check for balanced [ ]
    for ($i=0;$i<length($bfcode);$i++) {
        $c = substr($bfcode,$i,1);
        $nesting++ if ($c eq '[');
        $nesting-- if ($c eq ']');
    } 

#   A non-zero value in $nesting means we have unbalanced [ ]
    if ($nesting) {
        $msg = ($nesting>0?'[':']');
        warn "  Unbalanced brackets - too many $msg : conversion aborted\n";
    } else {
#       Write out comment block for the converted Brainfuck code
        $cvtcode .= "$cmt_open Generated by Brainfuck-to-Perl (bf2pl) $cmt_close$lineterm$lineterm";
        $cvtcode .= "$cmt_open Target language is $target $cmt_close$lineterm$lineterm";
        $cvtcode .= "$cmt_open Enabled options: $cmt_close$lineterm";
        if ($opt || $pchk || $pmax || $circ) {
            if ($classic) { $cvtcode .= "$cmt_open        Classic Mode $cmt_close$lineterm"; }
            if ($byte)    { $cvtcode .= "$cmt_open        Values limited to one byte $cmt_close$lineterm"; }
            if ($circ)    { $cvtcode .= "$cmt_open        Using circular array $cmt_close$lineterm"; }
            if ($opt)     { $cvtcode .= "$cmt_open        Repetition optimization $cmt_close$lineterm"; }
            if ($pchk)    { $cvtcode .= "$cmt_open        Pointer range checking $cmt_close$lineterm"; }
            if ($pmax)    { $cvtcode .= "$cmt_open        Max Pointer Size of $pmax $cmt_close$lineterm"; }
        } else {
            $cvtcode .= "$cmt_open        None $cmt_close";
        }
        $cvtcode .= "$lineterm$lineterm";

#       $prologue contains any startup / initialization code we may need
        if ($prologue ne "") {
            $cvtcode .= "$cmt_open".('-' x 60)."$cmt_close$lineterm";
            $cvtcode .= "$cmt_open Prologue:    $cmt_close$lineterm$lineterm";

            $cvtcode .= "$prologue$lineterm$lineterm";

            $cvtcode .= "$cmt_open".('-' x 60)."$cmt_close$lineterm$lineterm";
            $cvtcode .= "$cmt_open Main program:    $cmt_close$lineterm$lineterm";
        }

#       Here we go.  Process each Brainfuck instruction
        for ($i=0;$i<length($bfcode);$i++) {

            $prev_c = $c;
            $c = substr($bfcode,$i,1);

            for ($c) {

#               If rep optimization is enabled, this is where we write out
#               the optimized instruction.  If the current instruction
#               is not one of the inc/dec instructions, array dump or
#               same as the last instruction and optimization is enabled,
#               then print out the optimized instruction before processing
#               the current instruction.

                /([^\>\<\+\-\#]|[^$prev_c])/ && $opt && do {
                    unless ((/[\>\<]/ && ($prev_c =~ /[\>\<]/)) || (/[\+\-]/ && ($prev_c =~ /[\+\-]/))) {
                        $opt_code = '';
                        if ($cnt_p > 0) {
                            $cnt_p = abs($cnt_p);
                            if ($cnt_p == 1) { $opt_code = "$indstr$p_inc$lineterm"; }
                            else             { $opt_code = "$indstr$p_add$lineterm"; }
                            $opt_code =~ s/#cnt_p/$cnt_p/;
                            $cnt_p = 0;
                        }
                        if ($cnt_p < 0) {
                            $cnt_p = abs($cnt_p);
                            if ($cnt_p == 1) { $opt_code = "$indstr$p_dec$lineterm"; }
                            else             { $opt_code = "$indstr$p_sub$lineterm"; }
                            $opt_code =~ s/#cnt_p/$cnt_p/;
                            $cnt_p = 0;
                        }
                        if ($cnt_a > 0) {
                            $cnt_a = abs($cnt_a);
                            if ($cnt_a == 1) { $opt_code = "$indstr$ap_inc$lineterm"; }
                            else             { $opt_code = "$indstr$ap_add$lineterm"; }
                            $opt_code =~ s/#cnt_a/$cnt_a/;
                            $cnt_a = 0;
                        }
                        if ($cnt_a < 0) {
                            $cnt_a = abs($cnt_a);
                            if ($cnt_a == 1) { $opt_code = "$indstr$ap_dec$lineterm"; }
                            else             { $opt_code = "$indstr$ap_sub$lineterm"; }
                            $opt_code =~ s/#cnt_a/$cnt_a/;
                            $cnt_a = 0;
                        }
                        $cvtcode .= $opt_code;
                    }
                };

#               Process each instruction.

                /\>/ && do { 
                    if ($opt) { $cnt_p++; }
                    else      { $cvtcode .= "$indstr$p_inc$lineterm"; }
                    last;
                };
                /\</ && do { 
                    if ($opt) { $cnt_p--; }
                    else      { $cvtcode .= "$indstr$p_dec$lineterm"; }
                    last;
                };
                /\+/ && do { 
                    if ($opt) { $cnt_a++; }
                    else      { $cvtcode .= "$indstr$ap_inc$lineterm"; }
                    last;
                };
                /\-/ && do { 
                    if ($opt) { $cnt_a--; }
                    else      { $cvtcode .= "$indstr$ap_dec$lineterm"; }
                    last;
                };
                /\./ && do { 
                    $cvtcode .= "$indstr$putchar$lineterm";
                    last;
                };
                /\,/ && do { 
                    $cvtcode .= "$indstr$getchar$lineterm";
                    last;
                };
                /\[/ && do { 
                    $cvtcode .= "$indstr$while$lineterm";
                    $indlevel += $indamt;
                    $indstr = (' ' x $indlevel);
                    last;
                };
                /\]/ && do { 
                    $indlevel -= $indamt;
                    $indstr = (' ' x $indlevel);
                    $cvtcode .= "$indstr$endwhile$lineterm";
                    last;
                };
                /\#/ && do { 
                    $cvtcode .= "$indstr$ap_dump$lineterm";
                    last;
                };
                /\!/ && do { 
                    $cvtcode .= "$lineterm$indstr$exit $cmt_open ***** Forced Exit ***** $cmt_close$lineterm";
                    last;
                };
            }
        }
        $cvtcode .= "$lineterm";

#       $epilogue contains any cleanup code we may need
        if ($epilogue ne "") {
            $cvtcode .= "$cmt_open".('-' x 60)."$cmt_close$lineterm";
            $cvtcode .= "$cmt_open Epilogue:    $cmt_close$lineterm$lineterm";

            $cvtcode .= "$epilogue$lineterm$lineterm";

            $cvtcode .= "$cmt_open".('-' x 60)."$cmt_close$lineterm$lineterm";
        }
    }
    return $cvtcode;
}

#----------------------------------------------------------------------

sub DisplayHelp {

    print "bf2pl - Brainfuck to Perl converter.\n\n";
    print "Usage: bf2pl [options] filename [filename ...]\n\n";
    print "Options:\n\n";
    print "    -[no]byte    : Use single byte unsigned values.     (" . ($byte?'':'no')            ."byte)\n";
    print "    -[no]circ    : Use circular array of size pmax.     (" . ($circ?'':'no')            ."circ)\n";
    print "    -[no]classic : Sets -byte and -pmax=30000.          (" . ($classic?'':'no')         ."classic)\n";
    print "    -[no]exec    : Execute after conversion.            (" . ($exec?'':'no')            ."exec)\n";
    print "    -[no]opt     : Combine multiple inc/dec statements. (" . ($opt?'':'no')             ."opt)\n";
    print "    -[no]pchk    : Check for p < 0.                     (" . ($pchk?'':'no')            ."pchk)\n";
    print "    -pmax=val    : Sets max array size                  (" . ($pmax?"$pmax":"No Limit") . ")\n";
    print "    -target=lang : Sets the target language.            ($target)\n";
    print "    -help        : Display this message.\n";
    print "\n";

    exit;
}

#----------------------------------------------------------------------

sub Init_for_Perl {


#   File extention converted code
    $file_ext  = ".pl";

#   EOL terminator and indent amount
    $lineterm  = "\n";
    $indamt    = 4;

#   Comment delimiters
    $cmt_open  = "#";
    $cmt_close = "";

#   Startup / Cleanup code if needed

    $prologue .= "    \$pmax = $pmax;\n\n";
    $prologue .= "#   Initalize the array\n";
    $prologue .= "    for (\$p=0;\$p<\$pmax;\$p++) { \$a[\$p] = 0; }\n\n";
    $prologue .= "#   Reset \$p\n";
    $prologue .= "    \$p = 0;";
    $epilogue  = "";

#   Code to bail if p is less than zero
    $pchklow   = "";
    if ($pchk) {
        $pchklow = qq~ die "\\\$p (\$p) < 0 : \$!" if (\$p < 0);~;
    }

#   Code to bail if p is greater than the specified max array size
    $pchkhigh  = "";
    if ($pmax) {
        $pchkhigh = qq~ die "\\\$p (\$p) > \$pmax : \$!" if (\$p >= \$pmax);~;
    }

#   Code to wrap p around for circular array
    if ($circ) {
        $pchklow  = qq~ \$p = \$p % \$pmax;~;
        $pchkhigh = qq~ \$p = \$p % \$pmax;~;
    }

#   Code to enforce that all values are one byte unsigned values
    $bytechk   = "";
    if ($byte) {
        $bytechk = qq~ \$a[\$p] = \$a[\$p] % 256;~;
    }

#   Increment / Decrement p by one
    $p_inc    = "\$p++;$pchkhigh";
    $p_dec    = "\$p--;$pchklow";

#   Increment / Decrement p by more than one
    $p_add    = "\$p+=#cnt_p;$pchkhigh";
    $p_sub    = "\$p-=#cnt_p;$pchklow";

#   Increment / Decrement a[p] by one
    $ap_inc   = "\$a[\$p]++;$bytechk";
    $ap_dec   = "\$a[\$p]--;$bytechk";

#   Increment / Decrement a[p] by more than one
    $ap_add   = "\$a[\$p]+=#cnt_a;$bytechk";
    $ap_sub   = "\$a[\$p]-=#cnt_a;$bytechk";

#   Start / End a while loop
    $while    = "while (\$a[\$p]) {";
    $endwhile = "}";

#   Read / Write a character
    $getchar  = "\$a[\$p] = ord(getc);";
    $putchar  = "printf('%c',\$a[\$p]);";

#   Dump the first ten array elements
    $ap_dump  = qq~printf("\\n    %2d: %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d\\n    %*s\\n",scalar(\$p),\$a[0],\$a[1],\$a[2],\$a[3],\$a[4],\$a[5],\$a[6],\$a[7],\$a[8],\$a[9],3*\$p+6,"^");~;

#   Force the program to quit
    $exit     = "exit;";
}

#----------------------------------------------------------------------

